/*---------------------------------------------------------------
 * iservd.c
 * 
 * Computone Corporation IntelliServer daemon.
 *---------------------------------------------------------------
 */
/*
 * $RCSfile: iservd.c,v $
 * $Revision: 1.3 $
 * $Date: 1998/08/04 20:16:15 $
 *
 * $Header: //Iservd/iservd.c#3 $
 * 
 * $Log: iservd.c,v $
 * Revision 1.3  1998/08/04 20:16:15  chuckh
 * Removed all references to the locking mechanism. Also removed all boss
 * functionality.  This program now serves as the entity directly responsible
 * for communicating with an Intelliserver
 *
 * Revision 1.2  1998/07/31 21:30:47  chuckh
 * Changed version ID to reflect correct version
 *
 * Revision 1.1  1998/07/31 20:58:16  chuckh
 * Initial revision
 *
 */


/*---------------------------------------------------------------
 * Revision History:
 * 
 * 07/28/94	MAG	First draft
 * 09/15/94	MAG	Released for porting, iservd.doc written
 * 12/15/95	NSH	Numerous small changes, see comments
 * 05/23/96     NSH     Changes for RTS up/down and remote shells
 *                      for connections refused
 * 08/13/97     NSH     1.3.7
 *			Fixed very rare bug caused by -R option
 * 08/23/97     NSH     1.3.8
 *			Added -D option to remove our devices
 *			Fixed bug causing link when mknod wanted
 *---------------------------------------------------------------
 */

#define VERSION "2.2"

#include "isysinc.h"	/* System-dependent includes */

#include "iservopt.h"
#include "iservlog.h"
#include "iservcon.h"
#include "iservtel.h"
#include "iservpsu.h"


/*---------------------------------------------------------------
 * GLOBALS
 *---------------------------------------------------------------
 */

servboss_t Boss;	/* Grandfather information */
servopts_t Opts;	/* Command-line option information */
coninfo_t  Conn;	/* Connection informatin */
servpsu_t  Psu;		/* Psudo-tty information */

int Writer = 1;

int ReaderPid = 0;

char Me[128];

telbuf_t telbuf;	/* Buffer to pass to telnetEncode routine */

char rmt_cmd_flag = 0;	/* Bit flag for remote commands */
			/* should be using xxxFlag vars */

char rmt_dcd_flag = 0;	/* Boolean flag to indicate whether dcd is up or down */
char	absArgv0[MAXPATHLEN];

/*---------------------------------------------------------------
 * functions
 *---------------------------------------------------------------
 */

void mainBoss ARGS((servboss_t *));

static void lostPseudo ARGS(());
static void timeoutPseudo ARGS(());
static int gotTimeout ARGS(());
static void timeoutMsr ARGS(());



/*---------------------------------------------------------------
 * postStart(argc, argv)
 * Helper routine which catenates all the
 * command-line arguments together so 
 * they can be sent to the log file nicely.
 * NO check for overflow, but since the arguments
 * come from a config file, they are not likely to
 * be greater than 80... (not really)
 *---------------------------------------------------------------
 */

static
postStart(argc, argv)
	int argc;
	char ** argv;
{
	char catenateArgs[256];
	int i;

	memset(catenateArgs,0,sizeof catenateArgs);

	do_log(IL_NOTICE, "Main: Version %s Release %s", VERSION, RELEASE);

	for (i = 0; i < argc; i++)
	{
		strcat(catenateArgs, argv[i]);
		strcat(catenateArgs, " ");
	}

	if (strlen(catenateArgs) > 120)
	{
		strcpy(&catenateArgs[76], "...");
	}

	do_log(IL_NOTICE, catenateArgs);
	do_log(IL_NOTICE, "Parent pid %d", getpid());
}



/*---------------------------------------------------------------
 * dirtyExit(message)
 * called from several places in this module upon error
 *---------------------------------------------------------------
 */
static
dirtyExit(message)
char *message;
{
	do_log(IL_NOTICE, "Terminating: %s", message);
	if(ReaderPid)
	{
		int tmp = ReaderPid;
		/* No need to wait for child: I'm dying too */
		ReaderPid = 0;		/* so sigchild handler wont restart him */
		kill(tmp, SIGKILL);
	}
	cleanPseudo(&Psu);
	delPseudo(&Psu);
	exit(5);
}



/*---------------------------------------------------------------
 * signalExit
 * we get here from an "iservd -K"
 *---------------------------------------------------------------
 */		

static void
signalExit(sigtype)
int sigtype;
{
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);

	do_log(IL_NOTICE, "Pid %d terminating: signal %d", getpid(), sigtype);

	if(ReaderPid)
	{
		int tmp = ReaderPid;
		/* No need to wait for child: I'm dying too */
		ReaderPid = 0;
		do_log(IL_NOTICE, "SigKill pid %d", tmp);
		kill(tmp, SIGKILL);
	}

	cleanPseudo(&Psu);
	delPseudo(&Psu);
	exit(6);
}



/*---------------------------------------------------------------
 * mainChild(argc, argv, bos)
 * 
 * Main level for the top-level child process on each port.
 * It may be useful to thing of the lineage thusly:
 *   Grandparent process - the Boss that reads iservd.conf
 *   and starts a process for every line in that file then exits.
 *   Parent processes that send to the Iserver and fork readers.
 *   Children processes that read from the Iserver.
 * So this routine can be thought of as a "parent".
 *---------------------------------------------------------------
 */		

mainChild(argc, argv, bos)
int argc;
char ** argv;
servboss_t *bos;
{
	/* Check the command-line arguments to see if we have a chance */

	defaultOpts(&Opts);

	parseOpts(argc, argv, &Opts);

	if (rmt_cmd_flag & REMOTE_MSR_CMD)
		rmt_dcd_flag = 0;
	else
		rmt_dcd_flag = 1;

	rationalOpts(&Opts, bos);		/* opens lock file */

	if (Opts.svoFlag & SVO_SYNTAX)
	{
		fprintf(stderr, 
		  "%s called with bad arguments: check configuration file\n",
			absArgv0);
		fprintf(stderr, "Config error mesage: %s", Opts.svoError);
		exit(1);
	}

	/* Well, we are valid enough to try, anyway */

	setSignals();


	/* Well, we have the lock anyway */

	/* Make the standard input and the standard output harmless,
	 * If you are starting iservd from inittab you may want to
	 * use setpgrp to put iservd in a separate process group.
	 * We saw a strange problem on an HP machine that this fixed.
	 */

	/* setpgrp(); */

	close(0);
	close(1);

	open("/dev/null", O_RDONLY);
	open("/dev/null", O_WRONLY);

	/* Initialize any types of logging we are to use */

	if(!init_log(&Opts))
	{
		fprintf(stderr,"Cant init logging");
		exit(4);
	}

	/* Initialize other structures from command-line options */

	initConnect(&Conn, &Opts);
	initPsu(&Psu, &Opts);

	/* Log that we have started */
	postStart(argc, argv);

	/* Locate a master-slave pseudo pair */
	if (!(rmt_cmd_flag & REMOTE_MSR_CMD)) {
		if (!buildPseudo(&Psu)) 
			dirtyExit("Cannot create pseudo-device");
	}

	if (rmt_cmd_flag & REMOTE_MSR_CMD)
		intrSignal(SIGUSR2, timeoutMsr);

	intrSignal(SIGALRM, timeoutPseudo);

	signal(SIGINT,  signalExit);
	signal(SIGQUIT, signalExit);
	signal(SIGTERM, signalExit);

	do_log(IL_DEBUG,"hostToServer()");

	/* we will fork the reader process in this routine */

	for(;;)
	{
		hostToServer();
	}	
}



/*---------------------------------------------------------------
 * lostPseudo()
 * What to do when you lost the Pseudo-tty
 * connection, or it becomes idle. 
 * This is only called on the writer side;
 * no alarm is set on the reader side.
 *---------------------------------------------------------------
 */

static void
lostPseudo()
{
int status;

	if(!(Psu.psuFlag & PSU_SOPEN))
	{
		do_log(IL_NOTICE,"Slave port inactive");
	}
#ifndef SGI

	reopenSlave(&Psu);	/* Reopen slave port if closed */

#else
	do_log(IL_NOTICE,"SGI fix");

	if (!buildPseudo(&Psu))
	{
		dirtyExit("Cannot re-create pseudo-device");
	}
#endif
	if (isPermanent(&Conn)) return;

	endConnection(&Conn);

	if(ReaderPid)
	{
		int tmp = ReaderPid;
		do_log(IL_DEBUG,"Killing reader");
		ReaderPid = 0;
		kill(tmp, SIGKILL);
		wait(&status);	/* So death of child won't get us later */
	}	
}

int got_timeout = 0;

static void
timeoutPseudo()
{
	intrSignal(SIGALRM, timeoutPseudo);
	alarm(0);
	do_log(IL_NOTICE, "Slave port Timed out");
	got_timeout = 1;
}	

static int
gotTimeout()
{
int temp = got_timeout;

	got_timeout = 0;
	return temp;
}

int got_msr = 0;
static void
timeoutMsr()
{
	intrSignal(SIGUSR2, timeoutMsr);

	/* We get here if the child process(reader) has sent us this
	 * signal to indicate the dcd is up
	 */
	rmt_dcd_flag = 1;
	do_log(IL_DEBUG, "rcvd dcd up from child");
	got_msr = 1;
}

static int
gotMsr()
{
int temp = got_msr;

	got_msr = 0;
	return temp;
}


/*---------------------------------------------------------------
 * lostConnection()
 * you get here if you kill a port on the IntelliServer
 *---------------------------------------------------------------
 */

static void
lostConnection()
{
int status;
struct termio my_termio;

	do_log(IL_DEBUG,"Remote dropped connection");

	/* Mark the dcd flag to down if -M option was used */
	if (rmt_cmd_flag & REMOTE_MSR_CMD) {
		rmt_dcd_flag = 0;
		
		/* Reset CLOCAL here */
		if (!(Psu.psuFlag & PSU_SOPEN)) {
			do_log(IL_ERROR, "lostConnection: Slave not open!!");
		if ((Psu.psuSlave = open(Psu.psuSlaveName, O_RDWR)) < 0) {
			do_log(IL_ERROR, 
				"lostConnection: unable to open slave");
			goto wait_pid;
		}
			Psu.psuFlag |= PSU_SOPEN;
		}

		if (ioctl(Psu.psuSlave, TCGETA, &my_termio) < 0) {
			do_log(IL_ERROR,"TCGETA failed, error %d", errno);
			goto wait_pid;
		}

		my_termio.c_cflag &= ~CLOCAL;

		if (ioctl(Psu.psuSlave, TCSETA, &my_termio) < 0) 
			do_log(IL_ERROR,"TCSETA failed, error %d", errno);

	}
	else
		rmt_dcd_flag = 1;

wait_pid:
	if(ReaderPid)
	{
		int tmp = ReaderPid;
		ReaderPid = 0;
		kill(tmp, SIGKILL);
		wait(&status);	/* So death of child won't get us later */
	}

	markLost(&Conn, (char *) 0);
	endConnection(&Conn);

	/* If the appropriate option was given, loss of connection
	 * wants to be communicated to the slave side, by our 
	 * closing the master side. (Then we re-open and be ready
	 * for the next activation)
	 */

	if(Conn.conFlags & CON_HANG)
	{
		do_log(IL_NOTICE,"Slave port signalled");

		if (rmt_cmd_flag & REMOTE_MSR_CMD) {

			if (Psu.psuFlag & (PSU_SOPEN)) {
				Psu.psuFlag &= ~(PSU_SOPEN | PSU_CLEAN); 
				ClosePSU(Psu.psuSlave);
				/* close(Psu.psuSlave); */
				do_log(IL_NOTICE,"Slave closed. fd=%d",
					Psu.psuSlave);
			}
			cleanPseudo(&Psu); /* knock out any autopush before closing master */
			if (Psu.psuFlag & PSU_MOPEN) {
				Psu.psuFlag &= ~PSU_MOPEN;
				ClosePSU(Psu.psuMaster);
				/* close(Psu.psuMaster); */
				do_log(IL_NOTICE,"Master closed. fd=%d",
					Psu.psuMaster);
			}
		}
		else {
			if (!buildPseudo(&Psu))
			{
				dirtyExit("Cannot create pseudo-device");
			}
		}
	}
}



/*---------------------------------------------------------------
 * forkReader()
 * Attempts to start a "child" reader process,
 * Returns 0 if unsuccessful (a fatal condition)
 * Returns 1 otherwise.
 * Reads from the network/IntelliServer and writes to the pty
 *---------------------------------------------------------------
 */

int
forkReader()
{
int pid;
int status;

	do_log(IL_DEBUG,"forkReader()");

	if(ReaderPid)
	{
		int tmp = ReaderPid;
		ReaderPid = 0;
		kill(tmp, SIGKILL);
		wait(&status);	/* So we won't get his death later */
	}

	if ((pid = fork()) < 0) return 0;	/* Nasty error */

	if (pid > 0)
	{
		ReaderPid = pid;
		return 1;
	}

	do_log(IL_NOTICE, "Child pid %d", getpid());

	/* Then we are the child process */

	markChild();	/* Mark for subsequent log messages */

	Writer = 0;

	closeSlave(&Psu);
	alarm(0);

	for(;;)
	{
		serverToHost();
	}
}



/*---------------------------------------------------------------
 * connectAndFork()
 * called from hostToServer
 *---------------------------------------------------------------
 */

static int
connectAndFork()
{
	if (!attemptConnect(&Conn))
	{
		return 0;
	}
	/* Locate a master-slave pseudo pair */
	if (rmt_cmd_flag & REMOTE_MSR_CMD) {
		if (!buildPseudo(&Psu))
		{
			dirtyExit("Cannot create pseudo-device");
		}
	}

	if(!forkReader())
	{
		dirtyExit("Unable to fork reader process");
	}
#ifdef TYRED
	sleep(1);	/* Wait one second before returning from
			 * connection: this allows the server
			 * time to raise DTR, send his telnet
			 * replies, and so on. (Mostly of concern
			 * for temporary connections).
			 * (ideally we would wait until we received
			 *  negotiations from him, but that is not
			 *  practical using separate read and write
			 *  iservd processes.
			 */
#endif	/* TYRED */ /*DEBUGGING */
	telbuf.tb_state = TS_DATA;
	return 1;
}



/*---------------------------------------------------------------
 * serverToHost()
 * One pass of a perpetual loop to
 * service data transfer form server to host
 *---------------------------------------------------------------
 */

#define SHBUFFSIZE 1024

serverToHost()
{
char shBuffer[SHBUFFSIZE];
char shDecoded[SHBUFFSIZE];
int shCount;
int shWrote;
char * shOutput;

	shCount = read(Conn.conFd, shBuffer, sizeof(shBuffer));

	/* Any error reading the socket, and we will notify our
	 * parent through a SIGPIPE (so he knows he lost the connection)
	 * and we will exit.  A new reader will be spawned when the
	 * connection is re-established.
	 */

	if (shCount <= 0)
	{
		do_log(IL_NOTICE, "Child terminating: (net read returns %d, error %d)",
			shCount, errno);
		kill(getppid(), SIGPIPE);
		sleep(4);  /* allow time for PIPE to arrive before SIGCHILD */
		exit(1);
	}

	traceData("FromServer:", (unsigned char *)shBuffer, shCount);

	if(Conn.conFlags & CON_TELNET)
	{
		shOutput = shDecoded;
		shCount = telnetDecode((unsigned char *)shBuffer, 
				(unsigned char *) shDecoded, shCount, 
			Conn.conFlags & CON_BINARY,
			Conn.conFd);
	}
	else
	{
		shOutput = shBuffer;
	}

	while(shCount > 0)
	{
		shWrote = write(Psu.psuMaster, shOutput, shCount);

/* This amount of debugging is generally superfluous so we won't
 * compile it in unless you want it.
 */

#ifdef EXTRA_DEBUGGING
		traceData("ToPtyMstr:", (unsigned char *) shOutput,  shCount);
#endif /* EXTRA_DEBUGGING */

		if (shWrote <= 0)
		{
			/* EIO would indicate that the process using 
			 * the slave port has closed it.  So we would
			 * want our parent, the writer, to respond as
			 * though his pseudo read had timed out through
			 * an alarm.  If the connection is temporary, this
			 * may cause him to kill me, but otherwise, we will
			 * hang on.
			 * Then we try to give the writer plenty of time
			 * to get our signal and re-open the slave port,
			 * preventing further EIO's
			 */

			if (errno == EIO)
			{
				kill(getppid(), SIGALRM);
				sleep(2);
			}
			else	/* Actually, this would be unexpected */
			{
				sleep(1);
			}
		}
		else
		{
			shCount -= shWrote;
			shOutput += shWrote;
		}
	}
}



/*---------------------------------------------------------------
 * hostToServer()
 * One pass of a perpetual loop to service
 * data transfer from host to server.
 *---------------------------------------------------------------
 */

#define HSBUFFSIZE 1024

#define	IAC	255
hostToServer()
{
static char msr_query[3] = {IAC, MSR, 0};
char hsBuffer[HSBUFFSIZE];
char hsEncoded[HSBUFFSIZE * 2];
char do_seta=0;
int hsCount;
int hsWrote;
int hsError;
char * hsOutput;
struct termio my_termio;

	if(!isConnected(&Conn) && isPermanent(&Conn))
	{
		/* "legitimate" failure is possible here:
		 * user might have specified a connection retry
		 * limit for the permanent connection.  Ordinarily
		 * they would try forever, though.
		 */
		if(!connectAndFork())
		{
			return;
		}
	}

	/* check for DCD before writing, if DCD is down block the
	 * write here
	 */
	if (!rmt_dcd_flag)
		do_seta = 1;

	while (!rmt_dcd_flag) {
		do_log(IL_DEBUG, "rmt_dcd_flag is not set");
		sleep(Opts.svoMsrTimeout);
		/* query for MSR */
		if ((write(Conn.conFd, msr_query, 2) <= 0) ||
			gotpipe()) {
			do_log(IL_DEBUG, 
				"Got pipe after write %d", errno);
			lostConnection();
			return;
		}
		do_log(IL_DEBUG, "wrote msr_query to server");
	}

	
	if ((rmt_cmd_flag & REMOTE_MSR_CMD) && do_seta) {
		if (ioctl(Psu.psuSlave, TCGETA, &my_termio) < 0)
		{
			do_log(IL_ERROR,"TCGETA failed, error %d", errno);
			return;
		}
		my_termio.c_cflag |= CLOCAL;
		if (ioctl(Psu.psuSlave, TCSETA, &my_termio) < 0)
		{
			do_log(IL_ERROR,"TCSETA failed, error %d", errno);
			return;
		}
	}

	alarm(0);

	if(isConnected(&Conn))
	{
		gotTimeout();
		alarm(Opts.svoTimeout);
	}

	hsCount = read(Psu.psuMaster, hsBuffer, sizeof(hsBuffer));
	hsError = errno;

	alarm(0);

	if(gotpipe())
	{
		do_log(IL_DEBUG, "Got pipe after read %d", errno);
		lostConnection();
		return;
	}

	/* If hsCount < 0, we expect the error is probably EIO, or
	 * an EINTR from the alarm.
	 */

	if(hsCount <= 0)
	{
		/* Log any really strange errors as errors, and the
		 * kind we expect (when for example the port is closed
		 * as debugging output.
		 */

		if (hsCount < 0 && hsError == EINTR && gotMsr()) {
			do_log(IL_ERROR, "Master read %d, errno %d interrupted by reader process indicating the DCD value", hsCount, hsError);
			return;
		}
		else if(hsCount < 0 && hsError != EIO && !gotTimeout())
		{
			do_log(IL_ERROR,"Unexpected error %d reading master",
				hsError);
		}
		else
		{
			do_log(IL_DEBUG, "Master read %d, errno %d", 
				hsCount, hsError);
		}	
		lostPseudo();
		return;
	}

	/* Then there is data: */

	/* Since there is data, we should close our slave tty so that
	 * the user can have a last close, and therefore send us a
	 * signal when we are done.
	 */

	if (closeSlave(&Psu))
	{
		do_log(IL_NOTICE,"Slave port active");
	}

	/* If we are not already connected, we must attempt to connect.
	 * if attemptConnection fails, we have exhausted our retries,
	 * so we should cause the slave-device user to get an error on
	 * his write.  We do this by tearing down the master device
	 * and re-building it.
	 */

	if(!isConnected(&Conn))
	{
		if (!connectAndFork())
		{
			if (!buildPseudo(&Psu))
			{
				dirtyExit("Cannot create pseudo-device");
			}
			return;
		}
	}

	if(Conn.conFlags & CON_TELNET)
	{
		hsOutput = hsEncoded;
		telbuf.tb_buf = hsBuffer;
		hsCount = telnetEncode((unsigned char *)&telbuf, 
			(unsigned char *)hsEncoded, hsCount, 
			Conn.conFlags & CON_BINARY);
	}
	else
	{
		hsOutput = hsBuffer;
	}

	while(hsCount > 0)
	{

		hsWrote = write(Conn.conFd, hsOutput, hsCount);
		hsError = errno;

		if (hsWrote < 0 && hsError == EINTR && gotMsr()) {
			do_log(IL_ERROR, "Write %d, errno %d interrupted by reader process indicating the DCD value", hsCount, hsError);
			continue;
		}
		if (hsWrote <= 0 || gotpipe())
		{
			do_log(IL_DEBUG, "Got pipe after write %d", errno);
			lostConnection();
			return;
		}

		traceData("ToServer:", (unsigned char *) hsOutput,  hsWrote);
		hsCount -= hsWrote;
		hsOutput += hsWrote;
	}
}



/*---------------------------------------------------------------
 * usage
 * you can get this from invalid args like iservd -help
 *---------------------------------------------------------------
 */

void usage()
{
	fprintf(stderr,"iservd version %s release %s\n", VERSION, RELEASE);
	fprintf(stderr,"Usage:\n");
	fprintf(stderr,"iservd {Arguments} {Device} {Iserver} {Port}\n");
	fprintf(stderr,"	starts iservd using command line args.\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"	Arguments:\n");
	fprintf(stderr,"	-c	Initialize slave port to uncooked input\n");
	fprintf(stderr,"		but with cooked output\n");
	fprintf(stderr,"	-d	Seconds between retries: -d NNN\n");
	fprintf(stderr,"	-h	Hang up on disconnect\n");
	fprintf(stderr,"	-i	Ignore slave close (timeout only)\n");
	fprintf(stderr,"	-k	Name of lock file: -k AAA ***\n");
	fprintf(stderr,"		(not used in the config file)\n");
	fprintf(stderr,"	-l	Log file name: -l AAA\n");
	fprintf(stderr,"	-a	Disable binary mode\n");
	fprintf(stderr,"	-m	Disable binary mode (old-style)\n");
	fprintf(stderr,"	-o	Old iserved compat mode\n");
	fprintf(stderr,"	-p	Make permanent connection\n");
	fprintf(stderr,"	-r	Retry limit: -r NNN\n");
	fprintf(stderr,"	-s	Syslogging requested\n");
	fprintf(stderr,"	-t	Inactivity timeout: -t NNN\n");
	fprintf(stderr,"	-u	Initialize Slave port to uncooked mode\n");
	fprintf(stderr,"	-v	View data: -v [a|d|h]\n");
	fprintf(stderr,"	-w	Wait on close: -w NNN\n");
	fprintf(stderr,"	-x	Enable debugging output: -x\n");
	fprintf(stderr,"	-R	Enable the RTS command processing\n");
	fprintf(stderr,"	-I	Kill port on IS when connection is refused \n");
	fprintf(stderr,"\n");
	fprintf(stderr,"	Device:\n");
	fprintf(stderr,"		device node created in /dev\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"	Iserver:\n");
	fprintf(stderr,"		IP address or host name of IntelliServer\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"	Port:\n");
	fprintf(stderr,"		port number on IntelliServer + 9000\n");
}



/*---------------------------------------------------------------
 * MAIN: DISPATCH TO BOSS OR CHILD
 * The "grandfather" (Boss) iservd starts a child for every port.
 * These children become parents by starting a child reader.
 *---------------------------------------------------------------
 */

main(argc, argv)
int argc;
char ** argv;
{
	realpath(argv[0], absArgv0);

	/* initialize the data structure */
	defaultBoss(&Boss);

	/* We will always be a child */
	mainChild(argc, argv, &Boss);

	exit (0);
}


static void
mainBanner()
{
	fprintf(stderr,"\t------------------------------------\n");
	fprintf(stderr,"\t            iservd\n");
	fprintf(stderr,"\tVersion %s Release %s\n", VERSION, RELEASE);
	fprintf(stderr,"\tCompiled for %s\n", SPECIES);
	fprintf(stderr,"\tConf file: %s\n", Boss.bosConf);
	fprintf(stderr,"\tLock file: %s\n", Boss.bosLock);
	fprintf(stderr,"\tCopyright 1994,95,96,97 Computone\n");
	fprintf(stderr,"\t------------------------------------\n");
}
