/*--------------------------------
 * iservopt.h
 *-------------------------------
 */
/*
 * $RCSfile: iservopt.h,v $
 * $Revision: 1.1 $
 * $Date: 1998/07/31 21:12:44 $
 *
 * $Header: //Iservd/iservopt.h#1 $
 * 
 * $Log: iservopt.h,v $
 * Revision 1.1  1998/07/31 21:12:44  chuckh
 * Initial revision
 *
 */


/* servboss_t:
 * A place to put all the parent-based parameters which
 * are conditional on command-line arguments.
 */

#define BOS_MAX 80

typedef struct _servboss {
	int bosFlag;				/* Various booleans, see below */
	char bosConf[BOS_MAX];		/* Name of the iservd config file */
	char bosLock[BOS_MAX];		/* Name of the iservd "lock" file */
	char bosError[BOS_MAX+30];	/* Error message, if it exists */
	FILE *bosConfFile;			/* File pointer of fopened config file */
	int bosLockFd;				/* File descriptor of opened lock file */
} servboss_t;

#define BOS_CONF	1	/* Configuration file defined on command line */
#define BOS_LOCK	2	/* Lock file defined on command line */
#define BOS_KILL	4	/* Kill option selected: Only the lock file
						 * is used.  All processes listed in the
						 * lock file are killed.
						 */
#define BOS_ISLOCK	8	/* There IS a lock file opened */

/* servopts_t:
 * A place to put all the operational parameters which are
 * conditional upon command-line arguments. These are ones
 * which would obtain on a per-child basis.
 */

#define SV_MAX 80

typedef struct _servopts {
	int svoFlag;		/* Various booleans */
	int svoRetryDelay;	/* Seconds between connection retries */
	int svoRetryLimit;	/* Max number of connection retries */
	int svoTimeout;		/* No-traffic timeout for temp connections */
	int svoMsrTimeout;	/* Timeout between trying for dcd flag */
	int svoCloseDelay;	/* Number of seconds to wait when closing */
	int svoPort;		/* TCP port on the server to use */
	char svoLogFile[SV_MAX];	/* File name of log file */
	char svoLink[SV_MAX];		/* Name of the linked pseudo device */
	char svoServer[SV_MAX];		/* Name of the Intelliserver */
	char svoError[SV_MAX];		/* Error message while parsing */
	char svoLockFile[SV_MAX];	/* Lock file */
	char svoView;				/* View option used */
} servopts_t;

/* svoFlag bits: */

#define	SVO_BINARY	1		/* Binary mode on (default) */
#define SVO_OLD		2		/* Old Intelliserver compat. mode */
#define SVO_PERM	4		/* Permanent connection */
#define SVO_DEBUG	8		/* Debug */
#define SVO_SYNTAX	0x10	/* Option parser has seen a syntax error */
							/* (error message in svoError */
#define SVO_PORT	0x20	/* Port argument was given */
#define SVO_LINK	0x40	/* Link device name was given */
#define SVO_SERVE	0x80	/* Server name was given */
#define SVO_LOGOPT	0x100	/* Log file name specified */
#define SVO_VIEW	0x200	/* View option specified */
#define SVO_HANGUP	0x400	/* Hang up tty if tcp disconnects */
#define SVO_LOCK	0x800	/* Lock file specified */
#define SVO_RETRIES	0x1000	/* Retry option (-r) specified */
#define SVO_SYSLOG	0x2000	/* Syslog (-s) requested */
#define SVO_SRAW	0x4000	/* Initialize slave port to raw mode */
#define SVO_OCOOK	0x8000	/* Like SRAW, but with cooked output */
#define SVO_IGCLOSE	0x10000	/* For temporary connections, ignore slave close */
							/* Only disconnect on output timeout */
#define SVO_DELETE	0x20000	/* delete the device name linked to the slave */

#define SVO_REQUIRED (SVO_PORT | SVO_SERVE | SVO_LINK) 	/* The required ones */
#define SVO_LOG (SVO_DEBUG | SVO_LOGOPT)	/* Implies logfile needed */


#define IS_ARGS "acd:hik:l:mopr:st:M:uv:w:xRID"

/*------------------------------------------
 * Declarations for functs in iservopts.c
 *------------------------------------------
 */

extern int parseConfigLine ARGS((char *, char *));

#define PARSED_IGNORE	0
#define PARSED_ERROR	1
#define PARSED_GOOD	2

extern void parseOpts ARGS((int, char **, servopts_t *));

extern int parseBoss ARGS((int, char **, servboss_t *));

#define IS_BOSS		0
#define IS_CHILD	1
#define IS_FAILURE	2


#define IS_DEFAULT_CONF	"/etc/iservd.conf"
#define IS_DEFAULT_LOCK "/etc/iservd.pids"

/*
 * extern def of rmt_cmd_flag and  bit values :
 */
extern	char	rmt_cmd_flag;

#define	REMOTE_RTS_CMD		0x1
#define REMOTE_KILL_PORT	0x2
#define REMOTE_MSR_CMD		0x4

