/*-----------------------------------
 * isysinc.h
 *
 * system-dependent includes for any
 * and all modules in iservd tree.
 *-----------------------------------
 */
/*
 * $RCSfile: isysinc.h,v $
 * $Revision: 1.6 $
 * $Date: 1998/09/01 11:24:17 $
 *
 * $Header: //Iservd/isysinc.h#2 $
 * 
 * $Log: isysinc.h,v $
 * Revision 1.6  1998/09/01 11:24:17  chuckh
 * Defines a missing maxpathlen pp symbol. Also works around a
 * bug in realpath() in SCO 5.0.0
 *
 * Revision 1.5  1998/08/05 20:07:30  chuckh
 * Had to do some fiddling to get HP to define what I needed.
 * The key file was iservpsu.c
 *
 * Revision 1.4  1998/08/04 21:22:00  chuckh
 * Added include for MIPS
 *
 * Revision 1.3  1998/08/04 20:19:40  chuckh
 * Added an inlucde to support maximum path size limit.
 *
 * Revision 1.2  1998/08/03 13:42:02  chuckh
 * The GNU package placed termio in the main include directoy. So
 * the #include had to be changed from <sys/...> to just <...>
 *
 * Revision 1.1  1998/07/31 21:14:27  chuckh
 * Initial revision
 *
 */

/*--------------------- MIPS ------------------- */

#if defined(MIPS)

#define PICK_ONE_OS
#define ARGS(x)	x
#define SPECIES "MIPS" 

#include <sys/types.h>
#include <sys/termio.h>
#include <sys/sysmacros.h>

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <bsd43/netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <varargs.h>
#include <posix/sys/termios.h>

#endif

/*-------------------- HP ---------------------- */

#if defined (HP)

#define PICK_ONE_OS
#define ARGS(x)	()

#if defined(HPC)
#define SPECIES "HP/UX (ptym)"
#else
#define SPECIES "HP/UX"
#endif

#define	_INCLUDE_TERMIO
#include <termios.h>

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>

#include <varargs.h>

#endif

/*-------------------- SUN --------------------- */

#if defined (SUN)

#define PICK_ONE_OS
#define ARGS(x)	()
#define SPECIES "SunOS (sun4)"
#include <sys/types.h>

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <sys/termio.h>
#include <varargs.h>

#endif

/*--------------------- U7 ---------------- */

#if defined (U7)

#define PICK_ONE_OS
#define ARGS(x)	x
#define SPECIES "UnixWare 7"

#include <sys/types.h>
#include <ctype.h>
#include <sys/param.h>	
#include <unistd.h>

#include <stdio.h>
extern char *ptsname(int);	/* was not in stdio.h */
#include <fcntl.h>

#include <errno.h>
#include <sys/signal.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/termio.h>
#include <varargs.h>

#endif
/*--------------------- SVR4 ---------------- */

#if defined (SVR4)

#define PICK_ONE_OS
#define ARGS(x)	x
#define SPECIES "UNIX SVR4 (UnixWare, etc)"
#include <sys/types.h>

#include <stdio.h>
extern char *ptsname(int);	/* was not in stdio.h */
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>

#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <sys/termio.h>
#include <varargs.h>

#endif

/*--------------------- AIX32 ---------------- */

#if defined (AIX32)

#define PICK_ONE_OS
#define ARGS(x)	x
#define SPECIES "AIX (RS6000)"
#include <sys/types.h>

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <sys/termio.h>
#include <varargs.h>

#endif

/*------------------- IS_SCO ---------------- */

#if defined (IS_SCO)

#define PICK_ONE_OS
#define ARGS(x)	x
#define SPECIES "SCO UNIX"
#include <sys/types.h>

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <sys/netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/stropts.h>
extern char *ptsname();	/* Was not in unistd.h */
#include <errno.h>
#include <sys/termio.h>
#include <varargs.h>
/*
 * This MUST be defined as 4096 * due to a bug in realpath()
 */
#ifdef	MAXPATHLEN
#undef	MAXPATHLEN
#endif
#define	MAXPATHLEN	4096

#endif

/*------------------- IS_ISC ---------------- */

#if defined (IS_ISC)

#define PICK_ONE_OS

#define ARGS(x)	()
#define SPECIES "Interactive/AT&T UNIX (3.2)"
#include <sys/types.h>

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/stropts.h>
extern char *ptsname();	/* Was not in unistd.h */
#include <errno.h>
#include <net/errno.h>
#include <sys/termio.h>
#include <varargs.h>

#endif


/*--------------------- DGUX ---------------- */

#if defined (DGUX)

#define PICK_ONE_OS
#define ARGS(x)	x
#define SPECIES "DGUX AViiON"
#include <sys/types.h>

#include <stdio.h>
extern char *ptsname(int);	/* was not in stdio.h */
#include <ctype.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <sys/termio.h>
#include <varargs.h>

#endif

/*--------------------- LINUX ------------------- */
#if defined(LINUX)

#define PICK_ONE_OS
#define ARGS(x)	x
#define SPECIES "LINUX" 

#include <sys/types.h>

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/file.h>

#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <errno.h>
#include <termio.h>
#include <varargs.h>

#endif

#ifndef PICK_ONE_OS		/* MIPS doesn't like an "else" here! */

#include <Which/OS/Do/You/Want>	/* You didn't pick one on the list */

#endif


