/*
 *
 * $RCSfile: ss.c,v $
 * $Revision: 1.19 $
 * $Date: 1998/09/01 11:22:49 $
 *
 * Super server.  Starts up a number of children, waits for them
 * to die, then restarts them.  Also responsible for shutting all of
 * them down.
 *
 * If you're viewing this in vi, it's best to
 *	set tabstop=4
 *	set shiftwidth=4
 *
 * $Header: //Iservd/ss.c#3 $
 * 
 * $Log: ss.c,v $
 * Revision 1.19  1998/09/01 11:22:49  chuckh
 * Added missing define for sco. Also fixes a Memory Fault we were
 * getting with realpath() - a bug in the library.
 *
 * Revision 1.18  1998/08/25 19:26:45  chuckh
 * *** empty log message ***
 *
 * Revision 1.17  1998/08/25 13:16:28  chuckh
 * Removed calls to sigset within signal handlers.
 *
 * Revision 1.16  1998/08/25 13:15:44  chuckh
 * Minor changes
 *
 * Revision 1.15  1998/08/24 19:52:35  chuckh
 * NCR can't seem to handle sigaction as documented.  The older
 * sigset seems to work, but it leaves that nasty but really short
 * window at the front of signal catching functions.
 *
 * Revision 1.14  1998/08/19 14:45:36  chuckh
 * Made some changes that are specific to the SUN compilation.
 *
 * Revision 1.13  1998/08/10 15:30:58  chuckh
 * Didn't set some stuff for HP
 *
 * Revision 1.12  1998/08/07 19:55:41  chuckh
 * SIGUSR1 now increments the debug level wrapping from 0 through 3
 * SIGUSR2 now dumps the job list to the log.
 *
 * Revision 1.11  1998/08/07 14:59:28  chuckh
 * Fixed rcs headers
 *
 * Revision 1.10  1998/08/05 20:22:26  chuckh
 * Changed sigaction inits to be conditional upon HP being defined.
 *
 * Revision 1.9  1998/08/05 20:13:42  chuckh
 * Changed functions from ANSI to classic C. I also had problems
 * initializing sigaction structures statically - It seems HP has
 * implemented one of the fields as a union and classic C won't init
 * a union statically.  So now it's done a run time
 *
 * Revision 1.8  1998/08/04 20:04:50  chuckh
 * Now supports the -K and -R flags. -K will send a TERM to
 * an existing process by the same name. A -R will send a HUP to an
 * existing process by the same name.
 *
 * the dontRespawn flag is now used exclusively to prevent a job from
 * being respawned.  Death by signal other than TERM will set this flag.
 *
 * Some debugging code was added to the execvp() process. I found that
 * when execvp failed, I didn't have a clear indication why. On AIX, it was
 * found that the immediate death of a child didn't notify the parent. Thus
 * a sleep period was added after an execvp failure.
 *
 * Revision 1.7  1998/08/03 16:03:09  chuckh
 * Move debug level message below point where absolute path name is set.
 *
 * Revision 1.6  1998/08/03 15:43:30  chuckh
 * Removed Log() from this module and added code to detect the
 * proper termination of a child. exit(0) and SIGTERM are considered
 * normal. All others will prevent the child from being respawned
 * until the parent receives the next SIGHUP.
 *
 * Revision 1.5  1998/08/03 14:30:35  chuckh
 * absArgv0 is now global rather than static
 *
 * Revision 1.4  1998/07/31 20:53:30  chuckh
 * Cleaned up RCS headers
 *
 * Revision 1.3  1998/07/31 20:34:15  chuckh
 * Split NewJob function out of ss.c into a separate module
 *
 * Revision 1.2  1998/07/31 19:18:15  chuckh
 * Now removes pid file upon successful exit
 *
 * Revision 1.1  1998/07/31 19:12:30  chuckh
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#include <varargs.h>
#include <time.h>
#include <sys/param.h>

#define	SS_C
#include "ss.h"

/*
 * Some OS's (like SCO) don't define this, or it may be variable base
 * on the type of file being accessed. i.e. normal file, pipe, ...
 * For SCO, this MUST be defined at 4096 due to a bug in
 * realpath() in 5.0.0.  After that you may be able to reduce this
 * to a more reasonable value.
 */
#ifdef	IS_SCO
#ifdef	MAXPATHLEN
#undef	MAXPATHLEN
#endif
#define	MAXPATHLEN	4096
#endif

#ifndef	ARG_MAX
#define	ARG_MAX	1024
#endif

#define	CFG_FORMAT	"%s.conf"
#define	PID_FORMAT	"%s.pid"
#define	MOURNING	5	/* Time to wait after death of cld */

#define	DEBUG_VERBOSE	3
#define	DEBUG_NORMAL	2
#define	DEBUG_ERRORS	1

#define	min(x,y)	((x < y) ? (x) : (y))
#define	max(x,y)	((x > y) ? (x) : (y))

#define	DISTANT_FUTURE	(60L*60*24)	/* 1 day's worth of seconds */

extern	char	*optarg;
extern	int	optind;
extern	int	opterr;
/* extern	const	char	*sys_errlist[]; */

/* Variables defined in this module and used externally */
int	debug = 0;				/* Debugging defaults to off. */
char	*debugArg = "0";	/* Used only if debug != 0 */

/* Functions internal to this module */
void CatchTERM();
void CatchCLD();
void CatchHUP();
void CatchUSR1();
void CatchUSR2();
void Configure();
void ignore();	/* DMC*/
void Reconfigure();
JOB	*NewJob();

/* variables used in this module only */
static	char	rcsid[] = "$Id: //Iservd/ss.c#3 $";
static	int		shutdown = 0;
static	int		daemon = 1;	/* By default, we run in the background */
static	int		reconfig = 0;
static	time_t	now;
static	JOB		jobHead = {&jobHead, &jobHead};
static	char	configFile[MAXPATHLEN];
static	char	pidFile[MAXPATHLEN];
		char	absArgv0[MAXPATHLEN];
static	int		killMode = 0;

/* NCR does things a little differently... like not working with sigaction! */
#ifndef	NCR
/* Init in top of main() */
static	struct	sigaction	saCLD, saHUP, saTERM, saIGN, saUSR1, saUSR2;
#endif	/* ifndef NCR */

#ifdef	SUN
/*
 * Once again, SUN does things a little differently.
 */
char	*
strerror(err)
int	err;
{
	extern	char *sys_errlist[];
	extern	int	sys_nerr;
	static	char	buf[64];

	if (err > sys_nerr)
	{
		sprintf(buf, "Unknown error: %d", err);
		return (buf);
	}
	return (sys_errlist[err]);
}
#endif


int
main(argc, argv)
int	argc;
char	*argv[];
{
	int	i,c;

	strncpy(configFile, DEFAULT_CONF_FILE, sizeof(configFile)-1);

	opterr = 1;
	while ((c = getopt(argc, argv, "RKsf:d:")) != EOF)
	switch (c)
	{
	/* Turn debuging up a notch */
	case 'd':
		debug = atoi(optarg);
		debugArg = optarg;
		break;

	case 'f':
		strncpy(configFile, optarg, sizeof(configFile)-1);
		configFile[sizeof(configFile)-1] = 0;
		break;

	case 'K':
		killMode = SIGTERM;
		break;
		
	case 'R':
		killMode = SIGHUP;
		break;

	case 's':
		daemon = 0;
		break;

	case '?':
		printf("usage: %s [-sKR] [-d debugLevel] [-f fileName]\n", argv[0]);
		printf("where:\n");
		printf("\t-s forces %s to run foreground.\n", argv[0]);
		printf("\t-K sends signal TERM (terminate) to the currently running %s\n", argv[0]);
		printf("\t-R sends signal HUP (reconfig) to the currently running %s\n", argv[0]);
		printf("\t-f = Uses specified config file.\n");
		printf("\tdebugLevel produces output. 0=none, 1=errors only, 2=verbose. (0)\n");
		exit(1);
	}


	/*
	 * Get the real path of the invoked program
	 */
	if (realpath(argv[0], absArgv0))
	{
#if 0
		sprintf(configFile, CFG_FORMAT, absArgv0);
		Log(DEBUG_VERBOSE, "configFile=%s", configFile);
#endif
		sprintf(pidFile, PID_FORMAT, absArgv0);
		Log(DEBUG_VERBOSE, "pidFile=%s", pidFile);
	}
	else
	{
		fprintf(stderr, "Can't resolve path to %s: %s",
			argv[0], strerror(errno));
		exit(1);
	}

	/* Publish the current debug level. */
	if (debug >= DEBUG_NORMAL)
	{
		Log(debug, "Debug level set to %d", debug);
	}

	/*
	 * If we're in kill mode, find the pid of the possibly
	 * running super server.  This will be kept in the same
	 * pid file as what we've calculated above
	 */
	if (killMode)
	{
		FILE	*fp;
		pid_t	pid = 0;

		/* Open the pid file */
		if ((fp = fopen(pidFile, "r")) == (FILE *) 0)
		{
			Log(DEBUG_NORMAL, "%s: %s", pidFile, strerror(errno));
			exit(0);
		}

		/* Read the pid out of it */
		if (fscanf(fp, "%d", &pid) != 1)
		{
			Log(DEBUG_NORMAL, "Can't read %s: %s",
				pidFile, strerror(errno));
			fclose(fp);
			exit(0);
		}
		fclose(fp);

		Log(DEBUG_VERBOSE, "Sending process %d signal %d",
			pid, killMode);

		/* Send the signal */
		if (pid != 0)
			kill(pid, killMode);

		exit(0);
	}

	/*
	 * Do some initialization that *would* have been done if we
	 * had an ANSI compiler!
	 */
	saCLD.sa_sigaction = CatchCLD;
	sigemptyset(&saCLD.sa_mask);
	saCLD.sa_flags = 0;

	saHUP.sa_sigaction = CatchHUP;
	sigemptyset(&saHUP.sa_mask);
	saHUP.sa_flags = 0;

	saTERM.sa_sigaction = CatchTERM;
	sigemptyset(&saTERM.sa_mask);
	sigaddset(&saTERM.sa_mask, SIGTERM);
	sigaddset(&saTERM.sa_mask, SIGINT);
	saTERM.sa_flags = 0;

	saIGN.sa_sigaction = ignore;
	sigemptyset(&saIGN.sa_mask);
	saIGN.sa_flags = 0;

	saUSR1.sa_sigaction = CatchUSR1;
	sigemptyset(&saUSR1.sa_mask);
	saUSR1.sa_flags = 0;

	saUSR2.sa_sigaction = CatchUSR2;
	sigemptyset(&saUSR1.sa_mask);
	saUSR2.sa_flags = 0;
	/* Done... finally. */

	/*
	 * Setup to catch some signals
	 */
#ifdef	NCR
	sigset(SIGINT, CatchTERM);
	sigset(SIGTERM, CatchTERM);
	sigset(SIGHUP, CatchHUP);
	sigset(SIGQUIT, SIG_IGN);
	sigset(SIGALRM, SIG_IGN);
	sigset(SIGUSR1, CatchUSR1);
	sigset(SIGUSR2, CatchUSR2);
	sigset(SIGCLD, CatchCLD);
	sigset(SIGTTOU, SIG_IGN);
	sigset(SIGTTIN, SIG_IGN);
	sigset(SIGTSTP, SIG_IGN);
#else	/* ifdef NCR */
	 sigaction(SIGINT, &saTERM, (struct sigaction *) 0);
	 sigaction(SIGTERM, &saTERM, (struct sigaction *) 0);
	 sigaction(SIGHUP, &saHUP, (struct sigaction *) 0);
	 sigaction(SIGQUIT, &saIGN, (struct sigaction *) 0);
	 sigaction(SIGALRM, &saIGN, (struct sigaction *) 0);
	 sigaction(SIGUSR1, &saUSR1, (struct sigaction *) 0);
	 sigaction(SIGUSR2, &saUSR2, (struct sigaction *) 0);
	 sigaction(SIGCLD, &saCLD, (struct sigaction *) 0);
	 sigaction(SIGTTOU, &saIGN, (struct sigaction *) 0);
	 sigaction(SIGTTIN, &saIGN, (struct sigaction *) 0);
	 sigaction(SIGTSTP, &saIGN, (struct sigaction *) 0);
#endif	/* ifdef NCR else */

	/*
	 * Init the list of children we have to track
	 */
	Configure();
	if (jobHead.next == &jobHead)
	{
		printf("No commands to track were specified.\n");
		exit(1);
	}

	/*
	 * If this is a daemon process, force it into the background
	 */
	if (daemon)
	{
		pid_t	pid;

		for (i=0;i<NOFILE;i++)
			close(i);
		open("/dev/null",0/*O_RDONLY*/);	/* 0 = stdin */
		open("/dev/null",1/*O_RDONLY*/);	/* 1 = stdout */
		open("/dev/console",1/*O_RDONLY*/);	/* 2 = stderr */
		chdir("/");
		umask(0);

		if ((pid = fork()) < 0)
		{
			Log(DEBUG_ERRORS, "Couldn't fork to background: %s",
				strerror(errno));
			Log(DEBUG_ERRORS, "Running in foreground.");
			daemon = 0;	/* In case we reference it later */
		}
		else if (pid > 0)
		{
			/* We're the parent.  Exit. */
			exit(0);
		}
		/* The child lives on... */
		setpgrp();	/* no sigs from stdin */
	}


	/* Print a banner */
	Log(DEBUG_VERBOSE, "%s started as %d", argv[0], getpid());
	{
		FILE	*fp;

		if ((fp = fopen(pidFile, "w")) == (FILE *) 0)
		{
			Log(DEBUG_ERRORS, "Can't open pid file: %s: %s",
				pidFile, strerror(errno));
			exit(1);
		}
		if (fprintf(fp, "%d\n", getpid()) == 0)
		{
			Log(DEBUG_ERRORS, "Can't write to pid file %s: %s",
				pidFile, strerror(errno));
				fclose(fp);
				unlink(pidFile);
		}
		fclose(fp);
	}
		

	/* Look forever */
	while (1)
	{
		JOB	*job, *njob;
		int	later;

		now = time((time_t *) 0);

		/*
		 * Scan the list for children that have died, but don't
		 * need to be respawned.
		 */
		for (job = jobHead.next; job != &jobHead; job=njob)
		{
			njob = job->next;

			if (job->obsolete && job->pid == 0)
				FreeJob(job);
		}


		if (shutdown)
			later = CheckShutdown();
		else
		{
			if (reconfig)		/* set by a HUP */
				Reconfigure();	/* cleared by calling this. */
			later = RespawnChildren();
		}


		/*
		 * If we have something to do later, sleep until then.
		 */
		if (later)
		{
			sleep(later);
		}

		/*
		 * Otherwise, just wait for a signal.
		 */
		else
		{
			pause();
		}
	}
	/* unreachable */
}



int
CheckShutdown()
{
	JOB	*cur;
	int	offspringDead = 1;

	/*
	 * Search for children that are still alive
	 */
	for (cur = jobHead.next; cur != &jobHead; cur=cur->next)
	{
		if (cur->pid)
		{
			offspringDead = 0;
			break;
		}
	}

	/* If all of our children are dead, we can exit */
	if (offspringDead)
	{
		unlink(pidFile);
		Log(DEBUG_VERBOSE, "Done.");
		exit(0);
	}

	/* Otherwise, we'll wait a few seconds and check again */
	return (3);
}


/*
 *
 * By here, there may be children that need starting
 */
int
RespawnChildren()
{
	JOB	*cur;
	int	later = DISTANT_FUTURE;


	for (cur = jobHead.next; cur != &jobHead; cur=cur->next)
	{
		/*
		 * If the child is not running...
		 */
		if (cur->pid == 0)
		{
			pid_t	pid;

			/*
			 * If we're not supposed to respawn this one, don't.
			 *
			 * note that obsolete nodes will have been removed from
			 * the list by now.
			 */
			if (cur->dontRespawn)
				continue;

			/*
			 * ... And it's not yet time to respawn it...
			 */
			if (cur->timeOfDeath+MOURNING > now)
			{
				/* 
				 * It's not yet time to start again.
				 * Set an alarm for later.
				 */
				later = min(later, MOURNING);
				continue;
			}


			Log(DEBUG_VERBOSE, "Starting job %s", cur->id);

			/*
			 * Time to fork a child...
			 */
			if ((cur->pid = fork()) < 0)
			{
				cur->pid = 0;
				Log(DEBUG_ERRORS, "fork(): %s", strerror(errno));
				continue;
			}

			/*
			 * If we're the new child, exec to the
			 * appropriate program.
			 */
			if (cur->pid == 0)
			{
				char	*mp;
				int		err = errno;

				if (debug >= DEBUG_VERBOSE
				 && (mp = (char *) malloc(ARG_MAX)) == (char *) 0)
				{
					char	*cp;
					int		i;

					/*
					 * Why does SUN have to be so different with such a standard function?
					 * They've redefined sprintf to be
					 * char *sprintf(char *str, char *fmt, ...) where it always
					 * returns str.  Are they beginning to act a bit like Bill?
					 */
#ifdef	SUN
					cp = mp + strlen(sprintf(mp, "execvp("));
#else
					cp = mp+sprintf(mp, "execvp(");
#endif
					for (i = 0; cur->args[i]; ++i)
					{
#ifdef	SUN
						cp += strlen(sprintf(cp, "\"%s\", ", cur->args[i]));
#else
						cp += sprintf(cp, "\"%s\", ",
							cur->args[i]);
#endif
					}
#ifdef	SUN
					cp += strlen(sprintf(cp, ")"));
#else
					cp += sprintf(cp, ")");
#endif
					Log(DEBUG_VERBOSE, "%s", mp);
				}

				/* should never return */
				execvp(cur->args[0], cur->args);
				Log(DEBUG_ERRORS, "execvp(\"%s\"): %s",
					cur->args[0], strerror(errno));

				sleep(5);
				exit(1);
			}
		}
	}
	
	if (later == DISTANT_FUTURE)
		return (0);
	else
		return (later);
}

void
ignore(sig)
int	sig;
{
	return;
}

/*
 *
 * SIGCLD handling function
 *
 * This should mark the child as being dead
 *
 */
void
CatchCLD(sig)
int	sig;
{
	pid_t	deadChild;
	int	status;
	JOB	*cur;

	while (1)
	{
		switch (deadChild = waitpid(-1, &status, WNOHANG))
		{
		/* We have children, but no more have died. */
		case 0:
			break;

		/* Something went wrong */
		case -1:
			/* If we were interrupted try again */
			if (errno == EINTR)
				continue;

			/* We get ECHILD if there are no children left! */
			if (errno != ECHILD)
				Log(DEBUG_ERRORS, "waitpid(-1,,)=-1: %s",
					strerror(errno));
			break;

		/* Got a child */
		default:
			for (cur=jobHead.next; cur!=&jobHead; cur=cur->next)
			{
				if (deadChild == cur->pid)
				{
					cur->pid = 0;
					cur->timeOfDeath = time((time_t *) 0);

					/*
					 * Something serious may have gone wrong.
					 * If the child exited normally with a non-
					 * zero exit status, disable respawning of
					 * this child until the next reconfig.
					 * This would most likely occur due to an exec()
					 * failing, like file not found or something.
					 */
					if (WIFEXITED(status))
					{
						 if (WEXITSTATUS(status) != 0)
						 {
						 	Log(DEBUG_ERRORS,
								"Job %s exited normally with a status of %d",
								cur->id, WEXITSTATUS(status));
						}
					}
					else if (WIFSIGNALED(status))
					{
						if (WTERMSIG(status) != SIGTERM)
						{
						 	Log(DEBUG_ERRORS,
								"Job %s terminated due to receipt of a signal %d",
								cur->id, WTERMSIG(status));
							Log(DEBUG_ERRORS,
								"It will not be restarted until you do a \"kill -1 %d\"",
								getpid());
							cur->dontRespawn = 1;
						}
					}
					
					/* 
					 * We found the dead child. No need
					 * to keep looking...
					 */
					break;
				}	/* end of if pid found */
			}	/* end of for looking for child pid */

			/*
			 * There may be more than one dead child. Go back and
			 * look for more.
			 */
			continue;
		}	/* end of switch (waidpid) */

		/* If we break out of a case, also break out of forever */
		break;

	} /* end of while forever */

#ifndef	NCR
	/*
	 * Change the action of the signal back to us...
	 * and exiting this routine will re-enable the signal.
	 */
	sigaction(SIGCLD, &saCLD, (struct sigaction *) 0);
#endif
}


/*
 *
 * SIGTERM and SIGINT handling function
 *
 * This should set a flag and send a SIGTERM signal to
 * all the children.
 */
void
CatchTERM(sig)
int	sig;
{
	JOB	*cur;

	/*
	 * Comming in to this routine, the signal we received
	 * is blocked and set back to its default action.
	 */
	shutdown = 1;
	Log(DEBUG_NORMAL, "Signal %d caught. Killing children.", sig);

	for (cur = jobHead.next; cur != &jobHead; cur=cur->next)
		if (cur->pid)
			kill(cur->pid, SIGTERM);

	/*
	 * Change the action of the signal back to us...
	 * and exiting this routine will re-enable the signal.
	 */
#ifndef	NCR
	sigaction(sig, &saTERM, (struct sigaction *) 0);
#endif
}


/*
 *
 * This will catch a SIGHUP. It is used to reconfig by killing
 * children that are no longer listed, and establishing child nodes for
 * those that are new.
 */
void
CatchHUP(sig)
int	sig;
{
	reconfig = 1;

#ifndef	NCR
	sigaction(sig, &saHUP, (struct sigaction *) 0);
#endif
}


/*
 *
 * This will catch a SIGUSR1. It is used to bump up the debug level.
 * The debug level will increment on receipt of each signal reaching
 * the highest level of 3 before rotating back around to zero.
 */
void
CatchUSR1(sig)
int	sig;
{
	debug = (debug + 1) % 4;
	Log(debug, "Debug level set to %d", debug);

#ifndef	NCR
	sigaction(sig, &saUSR1, (struct sigaction *) 0);
#endif
}


/*
 *
 * This will catch a SIGUSR2. It is used to dump the child list to
 * the log file
 */
void
CatchUSR2(sig)
int	sig;
{
   JOB *job;
   char  *args;

   Log(debug, "Job List dump:");
   if ((args = (char *) malloc(ARG_MAX)) == (char *) 0) {
      Log(debug, "Can't dump jobs: %s", strerror(errno));
   }
   else
   {
           for (job = jobHead.next; job != &jobHead; job=job->next) {
              int i = 0;
              char   *cp = args;
              for (i = 0; job->args[i]; ++i) {
#ifdef	SUN
                 cp += strlen(sprintf(cp, "%s ", job->args[i]));
#else
                 cp += sprintf(cp, "%s ", job->args[i]);
#endif
              }
              Log(debug, "Job: %s: %s", job->id, args);
              Log(debug, "     PID:%d  Last time of death: %.24s",
                      job->pid, ctime(&job->timeOfDeath));
              Log(debug, "     Flags: %s %s",
                 job->dontRespawn ? "dontRespawn" : "",
                 job->obsolete ? "obsolete" : "");
           }
   }
#ifndef	NCR
	sigaction(sig, &saUSR2, (struct sigaction *) 0);
#endif
}


/*
 *
 *
 * Return a usable line from a config file.
 * Basically, all this does is filter out comments and
 * blank lines.
 */
char	*
GetConfigLine(fp)
FILE	*fp;
{
	static	char	buf[1024];
	char	*cp;

	while (fgets(buf, sizeof(buf), fp))
	{
		/* Strip out comments */
		if (buf[0] == '#')
			continue;
		if (buf[0] == '\n')
			continue;

		/*
		 * blank lines are not usable.
		 */
		if (strlen(buf) == 0)
			continue;

		return (buf);
	}
	return (char *) 0;
}


/*
 *
 * Setup each of the children.  This will usually be done
 * by reading some sort of configuration file.
 */
void
Configure()
{
	FILE	*fp;
	char	*buf;

	if ((fp = fopen(configFile, "r")) == (FILE *) 0)
	{
		Log(DEBUG_ERRORS, "%s: %s", configFile, strerror(errno));
		exit(1);
	}

	/*
	 * Get a line
	 */
	while (buf = GetConfigLine(fp))
	{
		JOB	*job;
		/*
		 * Create a child and
		 * link it in the list.
		 */
		if (job = NewJob(buf))
		{
			job->next = &jobHead;
			job->prev = jobHead.prev;
			job->prev->next = job->next->prev = job;
		} else {
			fprintf(stderr,"bad line %s in %s\n",buf,configFile);
		}
	}
	fclose(fp);
}


/*
 *
 * After receiving a SIGHUP, this program will re-read
 * the configuration file.  Any jobs that have changed arguments
 * will be killed and restarted with the new args.  Any new jobs
 # will be started (later by RespawnChildren), and any jobs that
 * have been removed from the configuration file will be killed
 * and removed from this list.
 */
void
Reconfigure()
{
	JOB		*job, *njob;
	FILE	*fp;
	char	*buf;

	reconfig = 0;

	/*
	 * Open the config file
	 */
	if ((fp = fopen(configFile, "r")) == (FILE *) 0)
	{
		Log(DEBUG_ERRORS, "%s: %s", configFile, strerror(errno));
		return;
	}


	/*
	 * Mark all the children as obsolete for now.
	 * Clear all dont respawn flags.
	 */
	for (job = jobHead.next; job != &jobHead; job=job->next)
	{
		job->obsolete = 1;
		job->dontRespawn = 0;
	}

	while (buf = GetConfigLine(fp))
	{
		JOB	*newJob;

		/* Couldn't parse it? skip it. */
		if ((newJob = NewJob(buf)) == (JOB *) 0)
			continue;

		/*
		 * Look for this job id in the existing list.
		 */
		for (job = jobHead.next; job != &jobHead; job=job->next)
		{
			if (strcmp(job->id, newJob->id) == 0)
			{
				int	argsChanged = 0;
				int	i;

				/* IT's there!
				 * Have any of the arguments changed?
				 */
				for (i = 0; job->args[i]; ++i)
				{
					/* Check for fewer arguments */
					if (newJob->args[i] == (char *) 0

					/* or a change of argument value */
					 || strcmp(job->args[i], newJob->args[i]))
					{
						argsChanged = 1;
						break;
					}
				}
				/*
				 * If everything this far matches, 
				 * check for additional arguments
				 */
				if (!argsChanged
				 && newJob->args[i] != (char *) 0)
				 	argsChanged = 1;

				/*
				 * If the arguments changed...
				 */
				if (argsChanged)
				{
					Log(DEBUG_VERBOSE, "Job %s has changed.", job->id);

					/*
					 * Copy some important infomation about restarting
					 * this job.
					 */
					newJob->pid = job->pid;
					newJob->timeOfDeath = job->timeOfDeath;

					/*
					 * Link the new node in PLACE of
					 * the exiting node.
					 */
					newJob->next = job->next;
					newJob->prev = job->prev;
					newJob->next->prev = newJob->prev->next = newJob;

					/*
					 * dontRespawn is false in the new node. Thus, killing the
					 * child will cause it to respawn a new one with the new
					 * arguments.  
					 */
					if (job->pid)
					{
						/*
						 * If this immediately causes a SIGCLD, the *new* node
						 * (since it has already been placed in the list)
						 * will be marked with pid=0, and later respawned
						 * with the new arguments by RespawnChildren.
						 */
						kill(job->pid, SIGTERM);
					}
					
					/* Get rid of the exiting node */
					job->next = job->prev = (JOB *) 0;
					FreeJob(job);
				}

				/*
				 * Since the arguments didn't change, we can
				 * free the new child, and allow the old one
				 * to respawn since we did find it in the list.
				 *
				 * Note that if the job is running, it continues to do
				 * so uninterrupted.
				 */
				else
				{
					FreeJob(newJob);
					job->obsolete = 0;
				}

				break;
			}
		}

		/*
		 * If we didn't find it in the exiting job list,
		 * it's a new job and needs to be spawned.
		 * Merely placing this in the list will suffice since
		 * RespawnChildren is called *after* we reconfigure.
		 */
		if (job == &jobHead)
		{
			Log(DEBUG_VERBOSE, "New job found: %s", newJob->id);
			newJob->next = &jobHead;
			newJob->prev = jobHead.prev;
			newJob->next->prev = newJob->prev->next = newJob;
		}
	}

	/*
	 * By now, all the lines have been read in.  We can
	 * close the file and scan the list for stuff that's
	 * still marked "obsolete".  Anything that is marked so
	 * should be killed and removed from the list after death.
	 */
	fclose(fp);

	for (job = jobHead.next; job != &jobHead; job=njob)
	{
		/*
		 * Record the location of the next node. This node
		 * may be deleted.
		 */
		njob = job->next;

	 	if (job->obsolete)
		{
			Log(DEBUG_VERBOSE, "Job %s has been removed", job->id);
			if (job->pid == 0)
				FreeJob(job);
			else
				kill(job->pid, SIGTERM);
		}
	}
}


/*
 *
 * Another handy free function.  This one frees a job node.
 */
void
FreeJob(cur)
JOB	*cur;
{
	int	argNo;


	/* Take it out of the list, if it's in there */
	if (cur->next && cur->prev)
	{
		cur->prev->next = cur->next;
		cur->next->prev = cur->prev;
	}

	/* Get rid of the ID */
	if (cur->id)
		free(cur->id);

	/* Get rid of each argument */
	if (cur->args)
	{
		for (argNo = 0; cur->args[argNo]; ++argNo)
			free(cur->args[argNo]);
		
		/* and the list container itself */
		free(cur->args);
	}

	/* Then finally, the node itself */
	free(cur);
}


/*
 *
 * Pretty simple. This will parse a possibly quoted string out of
 * a command line.
 *
 * It returns a pointer to a static area that gets overwritten with
 * each call.
 *
 * If cmd is non-null, starts parsing from the begining of the specified
 * string. If cmd is null, resumes parsing from the end of the previously
 * returned token. This means that the original location and contents
 * of cmd MUST BE PRESERVED BETWEEN CALLS.
 */
char	*
GetToken(cmd)
char	*cmd;
{
	static	char	*last = (char *) 0;
	static	char	tokBuf[4096];
	char	*d = tokBuf;
	int	quoteChar = 0;
	char	c;

	if (cmd)
		last = cmd;
	if (last == (char *) 0 || *last == 0)
		return (char *) 0;


	while (1) switch (c = *last++)
	{
	/* Handle the escape character */
	case '\\':
		switch (c = *last++)
		{
		case '\\':	*d++ = '\\';	break;
		case '\'':	*d++ = '\'';	break;
		case '\"':	*d++ = '\"';	break;
		case 'n':	*d++ = '\n';	break;
		case 't':	*d++ = '\t';	break;
		case 'f':	*d++ = '\f';	break;
		default:	*d++ = c;	break;
		}
		break;

	/* Single quotes */
	case '\'':
		switch (quoteChar)
		{
		/* If it's a close single quote */
		case '\'':
			quoteChar = 0;
			continue;
		/* If it's an open single quote */
		case 0:	
			quoteChar = c;
			continue;
		/* If it's a double quote embedded within a single quote */
		case '\"':
			*d++ = c;
			continue;
		}
		break;

	/* double quotes */
	case '\"':
		switch (quoteChar)
		{
		/* If it's a close double quote */
		case '\"':
			quoteChar = 0;
			continue;
		/* If it's an open double quote */
		case 0:	
			quoteChar = c;
			continue;
		/* If it's a single quote embedded within a double quote */
		case '\'':
			*d++ = c;
			continue;
		}
		break;

	/* End of token */
	case '\n':
	case ' ':
	case '\t':
	case 0:
		/*
		 * If we reaced the end of line, mark it so.
		 * otherwise, skip over white spaces
		 */
		if (c == '\n' || c == 0)
		{
			last = (char *) 0;
		}
		else
		{
			/*
			 * If we're in quoted text, just keep going 
			 */
			if (quoteChar)
			{
				*d++ = c;
				break;
			}

			/* Skip over whitespaces */
			while (*last && strchr(" \t\n", *last))
				++last;
		}


		/* Terminate the arg */
		*d++ = 0;
		return (tokBuf);

	/* Just copy a character */
	default:
		*d++ = c;
		break;
	}
	/* Never gets here */
}

